"""
Code created by Sorin Liviu Jurj for the paper called "Mobile Application for Receipt Fraud Detection Based on Optical Character Recognition".
More information about the paper can be found here: https://www.jurj.de/mobile-application-for-receipt-fraud-detection-based-on-optical-character-recognition/
"""
import numpy as np
import matplotlib.pyplot as plt
import cv2
import os
import keras
from tqdm import tqdm
import tensorflow as tf

def box2rects(contours):
    """
    Return the bounding rectangles of the input countours
    """
    return [cv2.boundingRect(c) for c in contours]

def extractCharacters(img):
    """
    Extract digits from the input image and returns the image and positions of each individual digit
    """
    original = img.copy()
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    height , width = img.shape
    img = cv2.bitwise_not(img)
    blurSize = 2*int(height*0.01/2)+1
    img = cv2.GaussianBlur(img, (blurSize, blurSize), 0)
    img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 2*int(height*0.9/2)+1, -20)
    contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    splits = []
    for x, y, w, h in box2rects(contours):
        if h>height*0.05 and w>width*0.02 and h<height*0.5 and w<width*0.2:
            split = original[max(0,y-10):y+h+20,max(0,x-10):x+w+20]
            splits.append((split, (x, y, w, h)))
    return splits

def cropped2digits(destination):
    """
    Takes cropped images located at data1/cropped and saves the extracted digits at the input destination
    """
    for fil in tqdm(os.listdir('data1/cropped')):
        pixels = cv2.imread(os.path.join('data1/cropped', fil))
        splits = extractCharacters(pixels)
        for split in splits:
            split = split[0]
            cv2.imwrite(os.path.join(destination, str(np.random.uniform(0,1000))+".jpg"), split)

def rotateRandomly(img):
    """
    Picks an angle between -10 and 10 and rotates the image by that angle
    """
    rows, cols = img.shape[0], img.shape[1]
    rotationAngle = np.random.uniform(-10, 10)
    # center of rotation, rotation angle, scale
    M = cv2.getRotationMatrix2D((cols/2, rows/2), rotationAngle, 0.9)
    return cv2.warpAffine(img,M,(cols,rows), borderMode=cv2.BORDER_REPLICATE)

def addShadow(img):
    """
    Chooses a random straigth line and shades one side of the image by a random amount
    """
    y1, y2 = np.random.randint(0, img.shape[0], 2)
    x1, x2 = np.random.randint(0, img.shape[1], 2)
    img = img.copy()
    if x1==x2:
        m = img.shape[0]*(y2-y1)
    else:
        m = (y2-y1)/(x2-x1)
    shade = np.random.randint(0, 30)
    sign = np.random.choice([-1,1]) 
    for y in range(img.shape[0]):
        for x in range(img.shape[1]):
            if y*sign > (m*x - m*x1 + y1)*sign:
                img[y][x] = img[y][x] - min(shade, img[y][x][0], img[y][x][1], img[y][x][2]) 
    return img

def randomDownWarp(img):
    """
    Warps the image randomly by enlarging the bottom side of the image
    """
    img = img.copy()
    height , width, _ = img.shape
    pad = np.random.randint(0, int(width*0.15))
    img = cv2.copyMakeBorder(img, 0, 0, pad, pad, cv2.BORDER_REPLICATE)
    finalH, finalW, _ = img.shape
    warpPoints = np.float32([ [0, 0], [pad, finalH], [finalW, 0], [finalW-pad, finalH] ])
    warpDest = np.float32([[0, 0], [0, finalH], [finalW, 0], [finalW, finalH]])
    M = cv2.getPerspectiveTransform(warpPoints, warpDest)
    img = cv2.warpPerspective(img, M, (finalW, finalH), borderMode=cv2.BORDER_REPLICATE )
    return img

def randomTopWarp(img):
    """
    Warps the image randomly by enlarging the top side of the image
    """
    img = img.copy()
    height , width, _ = img.shape
    pad = np.random.randint(0, int(width*0.15))
    img = cv2.copyMakeBorder(img, 0, 0, pad, pad, cv2.BORDER_REPLICATE)
    finalH, finalW, _ = img.shape
    warpPoints = np.float32([ [pad, 0], [0, finalH], [finalW-pad, 0], [finalW, finalH] ])
    warpDest = np.float32([[0, 0], [0, finalH], [finalW, 0], [finalW, finalH]])
    M = cv2.getPerspectiveTransform(warpPoints, warpDest)
    img = cv2.warpPerspective(img, M, (finalW, finalH), borderMode=cv2.BORDER_REPLICATE )
    return img

def randomLeftWarp(img):
    """
    Warps the image randomly by enlarging the left side of the image
    """
    img = img.copy()
    height , width, _ = img.shape
    pad = np.random.randint(0, int(height*0.15))
    img = cv2.copyMakeBorder(img, pad, pad, 0, 0, cv2.BORDER_REPLICATE)
    finalH, finalW, _ = img.shape
    warpPoints = np.float32([ [0, pad], [0, finalH-pad], [finalW, 0], [finalW, finalH] ])
    warpDest = np.float32([[0, 0], [0, finalH], [finalW, 0], [finalW, finalH]])
    M = cv2.getPerspectiveTransform(warpPoints, warpDest)
    img = cv2.warpPerspective(img, M, (finalW, finalH), borderMode=cv2.BORDER_REPLICATE )
    return img

def randomRightWarp(img):
    """
    Warps the image randomly by enlarging the right side of the image
    """
    img = img.copy()
    height , width, _ = img.shape
    pad = np.random.randint(0, int(height*0.15))
    img = cv2.copyMakeBorder(img, pad, pad, 0, 0, cv2.BORDER_REPLICATE)
    finalH, finalW, _ = img.shape
    warpPoints = np.float32([ [0, 0], [0, finalH], [finalW, pad], [finalW, finalH-pad] ])
    warpDest = np.float32([[0, 0], [0, finalH], [finalW, 0], [finalW, finalH]])
    M = cv2.getPerspectiveTransform(warpPoints, warpDest)
    img = cv2.warpPerspective(img, M, (finalW, finalH), borderMode=cv2.BORDER_REPLICATE )
    return img

def randomWarp(img):
    """
    Picks a random horizontal warp and a random vertical warp and applies both to the image
    """
    horizontalWarp = np.random.choice([randomLeftWarp, randomRightWarp])
    verticalWarp = np.random.choice([randomDownWarp, randomTopWarp])
    return verticalWarp(horizontalWarp(img))

def addRandomPadding(img):
    """
    Pads the image from each side by a random amount
    """
    height, width, _ = img.shape
    pad1 = np.random.randint(1, int(width*0.2))
    pad2 = np.random.randint(1, int(width*0.2))
    pad3 = np.random.randint(1, int(height*0.2))
    pad4 = np.random.randint(1, int(height*0.2))
    return cv2.copyMakeBorder(img, pad3, pad4, pad1, pad2, cv2.BORDER_REPLICATE)

def addPadding(img, pad1, pad2, pad3, pad4):
    """
    Add a padding to the input image according to the specified amounts
    """
    return cv2.copyMakeBorder(img, pad3, pad4, pad1, pad2, cv2.BORDER_REPLICATE)

def pad_and_resize(img, height, width):
    """
    Rezises and pad the image in order to return an image of the target height and width
    """
    img = img.copy()
    desiredRatio = height/width
    ratio = img.shape[0]/img.shape[1]
    if ratio == desiredRatio:
        return cv2.resize(img, dsize=(width, height))
    elif ratio > desiredRatio:
        img = cv2.resize(img, dsize=(max(1, int(height*img.shape[1]/img.shape[0])), height))
        delta_w = width - img.shape[1]
        img = cv2.copyMakeBorder(img, 0, 0, delta_w//2, delta_w-delta_w//2, cv2.BORDER_CONSTANT, value=[255, 255, 255])
        assert img.shape[0] == height
        assert img.shape[1] == width
        return img
    else:
        img = cv2.resize(img, dsize=(width, max(1,int(width*img.shape[0]/img.shape[1]))))
        delta_h = height - img.shape[0]
        img = cv2.copyMakeBorder(img, delta_h//2, delta_h-delta_h//2, 0, 0, cv2.BORDER_CONSTANT, value=[255, 255, 255])
        assert img.shape[0] == height
        assert img.shape[1] == width
        return img

def augmentThisData(data):
    """
    Applies the four random transformations to generate a new augmented sample
    """
    return addRandomPadding(rotateRandomly(addShadow(randomWarp(data))))

def dataAugmentation(source, destination):
    """
    From the source images augment the data into the destination
    """
    for digit in tqdm(os.listdir(source)):
        files = os.listdir(os.path.join(source, digit))
        for iteration in tqdm(range(1000)):
            originalImage = cv2.imread(os.path.join(source, digit, np.random.choice(files)))
            generatedImage = pad_and_resize(augmentThisData(originalImage), 100, 50)
            cv2.imwrite(os.path.join(destination, digit, "{}_{}.jpg".format(digit,iteration)), generatedImage)

def confusionMatrix(ytrue, ypred):
    """
    Confusion Matrix building function
    """
    ytrue = tf.argmax(ytrue, axis=1)
    ypred = tf.argmax(ypred, axis=1)
    return tf.confusion_matrix(ytrue, ypred, num_classes=12)

def recall(M, i):
    """
    Recall function
    """
    return M[i][i]/float((tf.reduce_sum(M, axis=0)[i]).eval())

def precision(M, i):
    """
    Precision function
    """
    return M[i][i]/float((tf.reduce_sum(M, axis=1)[i]).eval())

def F1Score(M, i):
    """
    F1Score function
    """
    prec = precision(M, i)
    recal = recall(M, i)
    return 2*precision(M, i)*recall(M, i)/(precision(M, i)+recall(M, i))

def shuffle_in_unison(a, b):
    """
    Shuffles a and b with the same permutations
    """
    assert len(a) == len(b)
    shuffled_a = np.empty(a.shape, dtype=a.dtype)
    shuffled_b = np.empty(b.shape, dtype=b.dtype)
    permutation = np.random.permutation(len(a))
    for old_index, new_index in enumerate(permutation):
        shuffled_a[new_index] = a[old_index]
        shuffled_b[new_index] = b[old_index]
    return shuffled_a, shuffled_b

def digitRecognizer():
    """
    Digit recognizer CNN model for the first algorithm
    """
    model = keras.models.Sequential()
    model.add(keras.layers.Conv2D(4, [10, 10],  input_shape=[100, 50, 1], strides=(2, 2), dilation_rate=(1, 1), activation='relu'))
    model.add(keras.layers.Dropout(0.5))
    model.add(keras.layers.MaxPooling2D(pool_size=(2, 2), strides=None, padding='valid', data_format=None))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(12, activation="relu"))
    model.add(keras.layers.Dense(12, activation="softmax"))
    model.compile(optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy'])
    return model
    
def plotHistory(history,name):
    """
    Plots a given training history from a keras model
    """
    acc = history.history['acc']
    val_acc = history.history['val_acc']
    loss = history.history['loss']
    val_loss = history.history['val_loss']
    epochs = range(1, len(acc)+1)
    with open("results_{}.csv".format(name), 'w') as w:
        w.write("Training accuracy,Validation accuracy,Training loss,Validation loss\n")
        for i in range(1, len(acc)):
            w.write("{},{},{},{}\n".format("%.4f" % acc[i], "%.4f" % val_acc[i], "%.4f" % loss[i], "%.4f" % val_loss[i]))
    plt.semilogx(epochs, acc, 'bo', color='green', label='Training acc')
    plt.semilogx(epochs, val_acc, 'b',color='blue',label='Validation acc')
    plt.xlabel("Epoch")
    plt.ylabel("Accuracy")
    plt.title('Training and validation accuracy')
    plt.legend()
    plt.figure()
    plt.semilogx(epochs, loss, 'bo',color='green', label='Training loss')
    plt.semilogx(epochs, val_loss, 'b',color='blue', label='Validation loss')
    plt.xlabel("Epoch")
    plt.ylabel("Loss")
    plt.title('Training and validation loss')
    plt.legend()
    plt.show()

def getTestData(dataLocation):
    """
    Loads test data from dataLocation and returns images and labels
    """
    targetHeight=100
    targetWidth=50
    labelsDict = {"zero":0, "one":1, "two":2, "three":3, "four":4, "five":5, "six":6, "seven":7, "eight":8, "nine":9, "comma": 10, "none":11}
    digits = os.listdir(dataLocation)
    images = []
    labels = []
    for digit in digits:
        selImages = os.listdir(os.path.join(dataLocation, digit))
        for chosenImage in selImages:
            img = cv2.imread(os.path.join(dataLocation, digit, chosenImage))
            height, width, _ = img.shape
            img = addPadding(img, int(width*0.1), int(width*0.1), int(height*0.1), int(height*0.1))
            img = pad_and_resize(img, targetHeight, targetWidth)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = np.array(img).reshape([targetHeight, targetWidth, 1])
            img = np.array(img, dtype=np.float64)
            mean = np.mean(img)
            std = np.std(img)
            img -= mean
            img /= std
            images.append(img)
            labels.append(digit)
    images = np.array(images)
    labels = [labelsDict[label] for label in labels]
    labels = np.array(labels)
    return images, labels

def getTestReceiptData(dataLocation):
    """
    Loads test receipt data from dataLocation and returns images and labels
    """
    height=20
    width=10
    labelsDict = {"zero":0, "one":1, "two":2, "three":3, "four":4, "five":5, "six":6, "seven":7, "eight":8, "nine":9, "comma": 10, "none":11}
    digits = os.listdir(dataLocation)
    images = []
    labels = []
    for digit in digits:
        selImages = os.listdir(os.path.join(dataLocation, digit))
        for chosenImage in selImages:
            img = cv2.imread(os.path.join(dataLocation, digit, chosenImage))
            img = pad_and_resize(img, height, width)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = np.array(img).reshape([height, width, 1])
            img = np.array(img, dtype=np.float64)
            mean = np.mean(img)
            std = np.std(img)
            img -= mean
            img /= std
            images.append(img)
            labels.append(digit)
    images = np.array(images)
    labels = [labelsDict[label] for label in labels]
    labels = np.array(labels)
    return images, labels

def train(source, destination,name="digits"):
    """
    Trains and saves the model with data from given source
    """
    model = digitRecognizer()
    images = []
    labels = []
    labelsDict = {"zero":0, "one":1, "two":2, "three":3, "four":4, "five":5, "six":6, "seven":7, "eight":8, "nine":9, "comma": 10, "none":11}
    for digit in os.listdir(source):
        label = digit
        for image in os.listdir(os.path.join(source, digit)):
            img = cv2.imread(os.path.join(source, digit, image))
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = np.array(img).reshape([100, 50, 1])
            img = np.array(img, dtype=np.float64)
            mean = np.mean(img)
            std = np.std(img)
            img -= mean
            img /= std
            images.append(img)
            labels.append(labelsDict[label])
    one_hot_labels = np.zeros((len(labels), 12))
    one_hot_labels[np.arange(len(labels)), labels] = 1
    one_hot_labels = np.array(one_hot_labels)
    images = np.array(images)
    images, one_hot_labels = shuffle_in_unison(images, one_hot_labels)
    history = model.fit(x=images, y=one_hot_labels, shuffle=True, batch_size=32, epochs=30, validation_split=0.1, callbacks=[keras.callbacks.TerminateOnNaN()])
    scores = model.evaluate(images,one_hot_labels,verbose=0)
    print("%s: %.2f%%" % (model.metrics_names[1],scores[1]*100))
    model.save_weights(destination)
    model.save("digit_model.h5")
    plotHistory(history,name)

def test(weightLocation, dataLocation):
    """
    Tests the model saved at weightLocation with data at dataLocation
    """
    model = digitRecognizer()
    model.load_weights(weightLocation)
    images, labels = getTestData(dataLocation)
    predictions = model.predict_on_batch(images)
    one_hot_labels = np.zeros((len(labels), 12))
    one_hot_labels[np.arange(len(labels)), labels] = 1
    M = confusionMatrix(one_hot_labels, predictions)
    classes = list(range(10))+["Comma", "Noise"]
    with open("results_digits_scores.csv", 'w') as w:
        with tf.Session().as_default():
            suma = 0
            M = M.eval()
            w.write("Digit,Test accuracy,Samples,Precision,Recall,F1Score\n")
            total = np.sum(M)
            for i in range(12):
                a=M[i][i]
                x=np.sum(M,axis=0)[i]
                y=np.sum(M,axis=1)[i]
                accuracy = float((total + 2*a - x - y)/float(total))
                suma += accuracy
                w.write("{},{},{},{},{},{}\n".format(classes[i],"%.4f" % accuracy, np.sum(M, axis=1)[i], "%.4f" % precision(M, i), "%.4f" % recall(M, i), "%.4f" % F1Score(M, i)))
            suma /= 12
            w.write("{},{}".format("Overall accuracy",suma))
            w.write("{},{}".format("Accuracy from confusion matrix",accuracyMat(M)))

def testReceipt(weightLocation, dataLocation,model):
    """
    Tests the receipt digit model saved at weightLocation with data an dataLocation
    """
    model.load_weights(weightLocation)
    
    images, labels = getTestReceiptData(dataLocation)
    
    predictions = model.predict_on_batch(images)
    one_hot_labels = np.zeros((len(labels), 12))
    one_hot_labels[np.arange(len(labels)), labels] = 1
    M = confusionMatrix(one_hot_labels, predictions)
    classes = list(range(10))+["Comma", "Noise"]
    with open("results_receipt_scores.csv", 'w') as w:
        with tf.Session().as_default():
            suma = 0
            M = M.eval()
            print("Confusion Matrix")
            print(M)
            w.write("Digit,Test accuracy,Samples,Precision,Recall,F1Score\n")
            total = np.sum(M)
            for i in range(12):
                a=M[i][i]
                x=np.sum(M,axis=0)[i]
                y=np.sum(M,axis=1)[i]
                accuracy = float((total + 2*a - x - y)/float(total))
                suma += accuracy
                w.write("{},{},{},{},{},{}\n".format(classes[i],"%.4f" % accuracy, np.sum(M, axis=1)[i], "%.4f" % precision(M, i), "%.4f" % recall(M, i), "%.4f" % F1Score(M, i)))
            suma /= 12
            w.write("{},{}".format("Overall accuracy",suma))
            w.write("{},{}".format("Accuracy from confusion matrix",accuracyMat(M)))

def show(img):
    """
    Function for showing images if cv2.imshow doesn't work
    """
    plt.imshow(img,  cmap = 'gray', interpolation = 'bicubic')
    plt.xticks([]), plt.yticks([])  # to hide tick values on X and Y axis
    plt.show()

def receiptDigitExtraction(img):
    """
    Extract digits from the input image and returns the image and positions of each individual digit
    """
    original = img.copy()
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    height , width = img.shape
    img = cv2.bitwise_not(img)
    blurSize = 2*int(width*0.01/2)+1
    img = cv2.GaussianBlur(img, (blurSize, blurSize), 0)
    img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 2*int(width*0.7/2)+1, -20)
    _, contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    splits = []
    for x, y, w, h in box2rects(contours):
        if h>height*0.05 and w>width*0.02:
            split = original[max(0,y-2):y+h+4,max(0,x-2):x+w+4]
            splits.append((split, (x, y, w, h)))

    return splits

def randomLightLevel(img):
    """
    Randomly changes the brightness of the image 
    """
    return cv2.add(img, np.array([np.random.uniform(-50, 50)]))

def receiptDataAugmentation(source, destination):
    """
    From the source receipt digits images from receipt augment the data into the destination
    """
    for digit in tqdm(os.listdir(source)):
        files = os.listdir(os.path.join(source, digit))
        for iteration in tqdm(range(100)):
            originalImage = cv2.imread(os.path.join(source, digit, np.random.choice(files)))
            generatedImage = pad_and_resize(augmentThisData(originalImage), 20, 10)
            cv2.imwrite(os.path.join(destination, digit, "{}_{}.jpg".format(digit,iteration)), generatedImage)

def receiptDigitRecognizer():
    """
    Digit recognizer CNN model for the receipt digits
    """
    model = keras.models.Sequential()
    model.add(keras.layers.Conv2D(20, [2, 2],  input_shape=[20, 10, 1], strides=(1, 1), padding='valid', data_format=None, dilation_rate=(1, 1), activation='relu'))
    model.add(keras.layers.MaxPooling2D(pool_size=(2, 2), strides=None, padding='valid', data_format=None))
    model.add(keras.layers.Dropout(0.5))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(20, activation="relu"))
    model.add(keras.layers.Dense(12, activation="softmax"))
    model.compile(optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy'])
    return model

def trainReceipt(source, model, destination,name="receipt"):
    """
    Trains and saves the receipt digit model with data from given source
    """
    images = []
    labels = []
    labelsDict = {"zero":0, "one":1, "two":2, "three":3, "four":4, "five":5, "six":6, "seven":7, "eight":8, "nine":9, "comma": 10, "none":11}
    for digit in os.listdir(source):
        label = digit
        for image in os.listdir(os.path.join(source, digit)):
            img = cv2.imread(os.path.join(source, digit, image))
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = cv2.resize(img,(20,10))
            img = np.array(img).reshape([20, 10, 1])
            img = np.array(img, dtype=np.float64)
            mean = np.mean(img)
            std = np.std(img)
            img -= mean
            img /= std
            images.append(img)
            labels.append(labelsDict[label])
    one_hot_labels = np.zeros((len(labels), 12))
    one_hot_labels[np.arange(len(labels)), labels] = 1
    one_hot_labels = np.array(one_hot_labels)
    images = np.array(images)
    images, one_hot_labels = shuffle_in_unison(images, one_hot_labels)
    history = model.fit(x=images, y=one_hot_labels, shuffle=True, batch_size=1000, epochs=1100, validation_split=0.1, callbacks=[keras.callbacks.TerminateOnNaN()])
    scores = model.evaluate(images,one_hot_labels,verbose=0)
    print("Model metric %s: %.2f%%" % (model.metrics_names[1],scores[1]*100))
    model.save_weights(destination)
    model.save("receipt_model.h5")
    plotHistory(history,name)

def train_test_CNN1():
    """
    Training and testing the first model
    """
    model = digitRecognizer()
    model.summary()
    cropped2digits("data/digitsFinal")
    dataAugmentation("data/digitsFinal", "data/augmentedFinal")
    weightLocation = "digit_weights.h5"
    dataLocation = "data/augmentedFinal"
    testLocation = "data/digitsFinal"
    train(dataLocation,weightLocation)
    test(weightLocation, testLocation)

def train_test_CNN2():
    """
    Training and testing the second model
    """
    receiptDataAugmentation('receiptDigits','augmentedReceipt')
    model = receiptDigitRecognizer()
    model.compile(optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy'])
    model.summary()
    weightLocation = 'receipt_weights.h5'
    dataLocation =   'receiptData/augmentedReceipt'
    testLocation = 'receiptData/receiptDigits'
    trainReceipt(dataLocation,model,weightLocation)
    testReceipt(weightLocation, testLocation,model)

def pickle_CNN1():
    """
    Serializing the first model's weights
    """
    model = digitRecognizer()
    weightLocation = 'digit_weights.h5'
    model.load_weights(weightLocation)
    layers = model.layers
    CNN1_conv2d_weights = layers[0].get_weights()[0]
    CNN1_conv2d_bias = layers[0].get_weights()[1]
    CNN1_dense1_weights = layers[4].get_weights()[0]
    CNN1_dense1_bias = layers[4].get_weights()[1]
    CNN1_dense2_weights = layers[5].get_weights()[0]
    CNN1_dense2_bias = layers[5].get_weights()[1]
    fnames = {"CNN1_conv2d_weights":CNN1_conv2d_weights,"CNN1_conv2d_bias":CNN1_conv2d_bias,
        "CNN1_dense1_weights":CNN1_dense1_weights,"CNN1_dense1_bias":CNN1_dense1_bias,
            "CNN1_dense2_weights":CNN1_dense2_weights,"CNN1_dense2_bias":CNN1_dense2_bias}
    for fname in fnames.keys():
        with open("{}.pickle".format(fname),"wb") as f:
            pickle.dump(fnames[fname],f)

def pickle_CNN2():
    """
    Serializing the second model's weights
    """
    model = receiptDigitRecognizer()
    weightLocation = 'receipt_weights.h5'
    model.load_weights(weightLocation)
    layers = model.layers
    CNN2_conv2d_weights = layers[0].get_weights()[0]
    CNN2_conv2d_bias = layers[0].get_weights()[1]
    CNN2_dense1_weights = layers[4].get_weights()[0]
    CNN2_dense1_bias = layers[4].get_weights()[1]
    CNN2_dense2_weights = layers[5].get_weights()[0]
    CNN2_dense2_bias = layers[5].get_weights()[1]
    fnames = {"CNN2_conv2d_weights":CNN2_conv2d_weights,"CNN2_conv2d_bias":CNN2_conv2d_bias,
        "CNN2_dense1_weights":CNN2_dense1_weights,"CNN2_dense1_bias":CNN2_dense1_bias,
            "CNN2_dense2_weights":CNN2_dense2_weights,"CNN2_dense2_bias":CNN2_dense2_bias}
    for fname in fnames.keys():
        with open("{}.pickle".format(fname),"wb") as f:
            pickle.dump(fnames[fname],f)

if __name__ == "__main__":
    train_test_CNN1()
    pickle_CNN1()
    train_test_CNN2()
    pickle_CNN2()
    

