"""
Code created by Sorin Liviu Jurj for the paper called "Mobile Application for Receipt Fraud Detection Based on Optical Character Recognition".
More information about the paper can be found here: https://www.jurj.de/mobile-application-for-receipt-fraud-detection-based-on-optical-character-recognition/
"""
import cv2
import time
import os
import keras
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt

def box2rects(contours):
    """
    Return the bounding rectangles of the input countours
    """
    return [cv2.boundingRect(c) for c in contours]

def addPadding(img, pad1, pad2, pad3, pad4):
    """
    Add a padding to the input image according to the specified amounts
    """
    return cv2.copyMakeBorder(img, pad3, pad4, pad1, pad2, cv2.BORDER_REPLICATE)

def pad_and_resize(img, height, width):
    """
    Rezises and pad the image in order to return an image of the target height and width
    """
    img = img.copy()
    desiredRatio = height/width
    ratio = img.shape[0]/img.shape[1]
    if ratio == desiredRatio:
        return cv2.resize(img, dsize=(width, height))
    elif ratio > desiredRatio:
        img = cv2.resize(img, dsize=(max(1, int(height*img.shape[1]/img.shape[0])), height))
        delta_w = width - img.shape[1]
        img = cv2.copyMakeBorder(img, 0, 0, delta_w//2, delta_w-delta_w//2, 
			cv2.BORDER_CONSTANT, value=[255, 255, 255])
        assert img.shape[0] == height
        assert img.shape[1] == width
        return img
    else:
        img = cv2.resize(img, dsize=(width, max(1,int(width*img.shape[0]/img.shape[1]))))
        delta_h = height - img.shape[0]
        img = cv2.copyMakeBorder(img, delta_h//2, delta_h-delta_h//2, 0, 0, 
			cv2.BORDER_CONSTANT, value=[255, 255, 255])
        assert img.shape[0] == height
        assert img.shape[1] == width
        return img

def digitRecognizer():
    """
    Digit recognizer CNN model
    """
    model = keras.models.Sequential()
    model.add(keras.layers.Conv2D(4, [10, 10],  input_shape=[100, 50, 1], strides=(2, 2), 
		padding='valid', data_format=None, dilation_rate=(1, 1), activation='relu', 
		use_bias=True, kernel_initializer='glorot_uniform', bias_initializer='zeros', 
		kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None, 
		kernel_constraint=None, bias_constraint=None))
    model.add(keras.layers.Dropout(0.5))
    model.add(keras.layers.MaxPooling2D(pool_size=(2, 2), strides=None, padding='valid', 
		data_format=None))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(12, activation="relu"))
    model.add(keras.layers.Dense(12, activation="softmax"))
    model.compile(optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy'])
    return model

def threshold(img):
	img = img.copy()
	img = img.astype('uint8')
	img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
	height , width = img.shape
	img = cv2.bitwise_not(img)
	blurSize = 2*int(height*0.01/2)+1
	threshParameter = 2*int(height*0.9/2)+1
	img = cv2.GaussianBlur(img, (blurSize, blurSize), 0)
	img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
		cv2.THRESH_BINARY, threshParameter, -20)
	return img

def findDigits(img):
	original = img.copy()
	height , width = img.shape[:2]
	img = threshold(img)
	_, contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
	splits = []
	for x, y, w, h in box2rects(contours):
		if h>height*0.05 and w>width*0.02:
			split = original[max(0,y-10):y+h+20,max(0,x-10):x+w+20]
			splits.append((split, (x, y, w, h)))
	return splits

def getPrice(img):
	splits = findDigits(img)
	if len(splits)<2:
		return 0.0
	splits.sort(key= lambda split: split[1][0])
	eur = []
	cents = []
	outputs = [recognizeDigit_digits(digit) for digit, _ in splits]
	comma = [output[0][10] if np.argmax(output)==10 else -1 for output in outputs]
	predictions = [ np.argmax(digit) for digit in outputs ]
	if len(comma)>0:
		i = np.argmax(comma)
		eur = [str(pred) for pred in predictions[:i] if (pred!=11 and pred!=10)]
		cents = [str(pred) for pred in predictions[i+1:] if (pred!=11 and pred!=10)]
		try:
			return float("{}.{}".format("".join(eur), "".join(cents)))
		except:
			return 0.0
	else:
		eur = [str(pred) for pred in predictions if (pred!=11 and pred!=10)]
		try:
			return float("".join(eur))
		except:
			return 0.0

def recognizeDigit_digits(img):
	img = img.copy()
	img = img.astype('uint8')
	img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
	width, height = img.shape[:2]
	img = addPadding(img, int(width*0.1), int(width*0.1), int(height*0.1), int(height*0.1))
	img = pad_and_resize(img, 100, 50)
	targetHeight = 100
	targetWidth = 50
	img = np.array(img).reshape([-1, targetHeight, targetWidth, 1])
	img = np.array(img, dtype=np.float64)
	img -= np.mean(img)
	img /= np.std(img)
	return model_digits.predict(img)

def show(img):
	plt.imshow(img,  cmap = 'gray', interpolation = 'bicubic')
	plt.xticks([]), plt.yticks([])  # to hide tick values on X and Y axis
	plt.show()

def result_model_digits():
	directory =  'data/cropped'
	with open("results_model_digits_scores.csv", 'w') as w:
		w.write("Number;CNN_thresh;thresh_score\n")
		score_th = 0
		score_n_th = 0
		total = 0.0
		suma = 0.0
		nr = 0.0
		for filename in os.listdir(directory):
			total += 1
			img = cv2.imread(os.path.join(directory, filename))
			fname = filename.split(".")[0]
			fname = fname.split("_")
			fn = "{}.{}".format(fname[0],fname[1])
			start = time.time()
			text_no_thresh = getPrice(img)
			stop = time.time()
			suma += stop-start
			nr += 1

			if float(fn) != float(text_no_thresh):
				score_n_th += 1
				no_thresh_score = 1
			else :
				no_thresh_score = 0
			w.write("{};{};{}\n".format(fn,text_no_thresh,no_thresh_score))
		print("Average time {} sum {} number {}".format(suma/nr,suma,nr))
		w.write("{};{};{}\n".format("","WRONG",score_n_th))
		w.write("{};{};{}\n".format("","WRONG",(score_n_th/total)*100))
		w.write("{};{};{}\n".format("","RIGHT",100-score_n_th/total*100))

def threshold_receipt(img):
	original = img.copy()
	img = img.copy()
	img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
	height , width = img.shape
	img = cv2.bitwise_not(img)
	blurSize = 2*int(width*0.01/2)+1
	threshParameter = 2*int(width*0.7/2)+1
	img = cv2.GaussianBlur(img, (blurSize, blurSize), 0)
	img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, threshParameter, -20)
	return img

def findDigits_receipt(img):
	original = img.copy()
	height , width = img.shape[:2]
	img = threshold_receipt(img)
	_, contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
	splits = []
	for x, y, w, h in box2rects(contours):
		split = original[max(0,y-2):y+h+4,max(0,x-2):x+w+4]
		splits.append((split, (x, y, w, h)))
	return splits

def splitRows(img):
	original = img.copy()
	img = threshold_receipt(img)
	height, width = img.shape
	splits = []
	for i in range(height):
		if all([pixel==0 for pixel in img[i]]):
			splits.append(i)
	i = 0
	rows = []
	while i<height:
		while i in splits:
			i+=1
		if i < height:
			rows.append([])
		while i not in splits and i < height:
			rows[-1].append(original[i])
			i+=1
	rows = [np.array(row) for row in rows]
	return rows

def recognizeDigit_receipt(img):
	img = img.copy()
	width, height = img.shape[:2]
	img = pad_and_resize(img, 20, 10)
	img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
	targetHeight = 20
	targetWidth = 10
	img = np.array(img).reshape([-1, targetHeight, targetWidth, 1])
	img = np.array(img, dtype=np.float64)
	img -= np.mean(img)
	img /= np.std(img)
	return model_receipt.predict(img)

def getReceiptPrices(img):
	rows = splitRows(img)
	prices = []
	for row in rows:
		splits = findDigits_receipt(row)
		if len(splits)<1:
			continue
		splits.sort(key= lambda split: split[1][0])
		eur = []
		cents = []
		outputs = [recognizeDigit_receipt(digit) for digit, _ in splits]
		comma = [output[0][10] if np.argmax(output)==10 else -1 for output in outputs]
		predictions = [ np.argmax(digit) for digit in outputs ]
		if len(comma)>0:
			i = np.argmax(comma)
			eur = [str(pred) for pred in predictions[:i] if pred!=11 and pred!=10]
			cents = [str(pred) for pred in predictions[i+1:] if pred!=11 and pred!=10]
			try:
				prices.append(float("{}.{}".format("".join(eur), "".join(cents))))
			except Exception as e:
				prices.append(0.0)
				print(e)
		else:
			eur = [str(pred) for pred in predictions if pred!=11 and pred!=10]
			try:
				prices.append(float("".join(eur)))
			except:
				prices.append(0.0)
	return prices

def receiptDigitRecognizer():
    """
    Digit recognizer CNN model for the receipt digits
    """
    model = keras.models.Sequential()
    model.add(keras.layers.Conv2D(20, [2, 2],  input_shape=[20, 10, 1], strides=(1, 1), padding='valid', data_format=None, dilation_rate=(1, 1), activation='relu', use_bias=True, kernel_initializer='glorot_uniform', bias_initializer='zeros', kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None, kernel_constraint=None, bias_constraint=None))
    model.add(keras.layers.MaxPooling2D(pool_size=(2, 2), strides=None, padding='valid', data_format=None))
    model.add(keras.layers.Dropout(0.5))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(20, activation="relu"))
    model.add(keras.layers.Dense(12, activation="softmax"))
    model.compile(optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy'])
    return model

def result_model_receipt():
	directory =  'receiptData/photos'
	with open("results_model_scores_receipt.csv", 'w') as w:
		w.write("Number;CNN\n")
		suma = 0.0
		nr = 0.0
		for filename in os.listdir(directory):
			img = cv2.imread(os.path.join(directory, filename))
			fname = filename.split(".")[0]
			start = time.time()
			text = getReceiptPrices(img)
			stop = time.time()
			suma += stop-start
			nr += 1
			w.write("{};{}\n".format(fname,text))
		print("Average time {} sum {} number {}".format(suma/nr,suma,nr))

model_digits = digitRecognizer()
model_receipt = receiptDigitRecognizer()

def test_model_digits():
	weightLocation = "digit_weights.h5"
	model_digits.load_weights(weightLocation)
	model_digits.summary()
	start = time.time()
	result_model_digits()
	end = time.time()
	print(end - start)

def test_model_receipt():
	weightLocation = "receipt_weights.h5"
	model_receipt.load_weights(weightLocation)
	model_receipt.summary()
	start = time.time()
	result_model_receipt()
	end = time.time()
	print(end - start)

test_model_digits()
test_model_receipt()
